/*
 *  nploader module
 *
 *  Copyright (C) 2011  Codestation
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <pspsdk.h>
#include <pspiofilemgr.h>
#include <string.h>
#include <inttypes.h>
#include "nploader.h"
#include "hook.h"
#include "logger.h"

#define PSP_O_NPDRM 0x40000000

u32 sctrlHENFindFunction(char *modname, char *libname, u32 nid);

int decrypted = 0;

SceUID (*sceIoOpen_func)(const char *file, int flags, SceMode mode) = NULL;
SceUID (*sceIoOpenAsync_func)(const char *file, int flags, SceMode mode) = NULL;
int (*sceNpDrmEdataSetupKey_func)(SceUID fd) = NULL;

SceOff (*sceNpDrmEdataGetDataSize_func)(SceUID fd) = NULL;
int (*sceNpDrmRenameCheck_func)(const char *file) = NULL;

SceUID (*sceKernelLoadModule2_func)(const char *path, int flags, SceKernelLMOption *option) = NULL;
SceUID (*sceKernelLoadModule_func)(const char *path, int flags, SceKernelLMOption *option) = NULL;

SceOff (*sceIoLseek_func)(SceUID fd, SceOff offset, int whence) = NULL;

// the games calls this function (sceNpDrmEdataSetupKey) before reading the edat files
// returns 0 on success or a SCE error code otherwise
int np_setup(SceUID fd) {
    int k1 = pspSdkSetK1(0);
    kprintf("sceNpDrmEdataSetupKey called with fd: %08X\n", fd);
    pspSdkSetK1(k1);
    if (decrypted) {
        pspSdkSetK1(0);
        kprintf("DLC already decrypted for fd: %08X, returning 0 for sceNpDrmEdataSetupKey\n", fd);
        pspSdkSetK1(k1);
        return 0;
    }
    if (!sceNpDrmEdataSetupKey_func) {
        pspSdkSetK1(0);
        sceNpDrmEdataSetupKey_func = (void *) sctrlHENFindFunction("scePspNpDrm_Driver", "scePspNpDrm_user", 0x08D98894);
        //kprintf("Found sceNpDrmEdataSetupKey addr: %08X\n", (u32)sceNpDrmEdataSetupKey_func);
        pspSdkSetK1(k1);
    }
    int res = sceNpDrmEdataSetupKey_func(fd);
    k1 = pspSdkSetK1(0);
    kprintf("sceNpDrmEdataSetupKey returned: %08X\n", res);
    pspSdkSetK1(k1);
    return res;
}

// check if the file is encrypted with the edat format
int check_encryption(const char *file) {
    SceUID fd = sceIoOpen(file, PSP_O_RDONLY, 0777);
    kprintf("Checking for encryption: %s, fd: %08X\n", file, fd);
    if (fd >= 0) {
        char buffer[8];
        int read = sceIoRead(fd, buffer, sizeof(buffer));
        sceIoClose(fd);
        if (read == sizeof(buffer)) {
            if (memcmp(buffer, "\0PSPEDAT", sizeof(buffer))) {
                return 0;
            } else {
                return 1;
            }
        } else {
            return read;
        }
    }
    return fd;
}

// generic function for both sceIoOpen/sceIoOpenAsync
SceUID _npprx_open(const char *file, int flags, SceMode mode, int async) {
    // only check for files with the npdrm flag on that doesn't came from the UMD
    if (((flags & PSP_O_NPDRM) == PSP_O_NPDRM) && (memcmp(file, "disc0", 5)
            && memcmp(file, "DISC0", 5))) {
        int k1 = pspSdkSetK1(0);
        int res = check_encryption(file);
        switch (res) {
        case 0: //DECRYPTED
            flags &= ~PSP_O_NPDRM;
            kprintf("> Already decrypted\n");
            decrypted = 1;
            break;
        case 1: //ENCRYPTED
            kprintf("> Encrypted file\n");
            decrypted = 0;
            break;
        default:
            kprintf("Failed to open/read %s, code: %08X\n", file, res);
        }
        pspSdkSetK1(k1);
    }
    return async ? sceIoOpenAsync_func(file, flags, mode) : sceIoOpen_func(
            file, flags, mode);
}

// sceIoOpen hook function
SceUID np_open(const char *file, int flags, SceMode mode) {
    int k1;
    if (!sceIoOpen_func) {
        k1 = pspSdkSetK1(0);
        sceIoOpen_func = (void *) sctrlHENFindFunction("sceIOFileManager", "IoFileMgrForUser", 0x109F50BC);
        kprintf("Found sceIoOpen addr: %08X\n", (u32)sceIoOpen_func);
        pspSdkSetK1(k1);
    }
    int res = _npprx_open(file, flags, mode, 0);
    k1 = pspSdkSetK1(0);
    kprintf("sceIoOpen called: %s, flags: %08X, mode: %08X, fd: %08X\n", file, flags, mode, res);
    pspSdkSetK1(k1);
    return res;
}

// sceIoOpenAsync hook function
SceUID np_open_async(const char *file, int flags, SceMode mode) {
    int k1 = pspSdkSetK1(0);
    if (!sceIoOpenAsync_func) {
        sceIoOpenAsync_func = (void *) sctrlHENFindFunction("sceIOFileManager", "IoFileMgrForUser", 0x89AA9906);
        kprintf("Found sceIoOpenAsync addr: %08X\n", (u32)sceIoOpenAsync_func);
    }
    kprintf("sceIoOpenAsync called: %s, flags: %08X, mode: %08X\n", file, flags, mode);
    pspSdkSetK1(k1);
    return _npprx_open(file, flags, mode, 1);
}

// returns the real size of the decrypted edat file (sceNpDrmEdataGetDataSize)
SceOff np_size(SceUID fd) {
    int k1 = pspSdkSetK1(0);
    if (!sceNpDrmEdataGetDataSize_func) {
        sceNpDrmEdataGetDataSize_func = (void *) sctrlHENFindFunction("scePspNpDrm_Driver", "scePspNpDrm_user", 0x219EF5CC);
        //kprintf("sceNpDrmEdataGetDataSize addr: %08X\n", (u32)sceNpDrmEdataGetDataSize_func);
    }
    if (decrypted) {
        if (!sceIoLseek_func) {
            sceIoLseek_func = (void *) sctrlHENFindFunction("sceIOFileManager","IoFileMgrForUser", 0x27EB27B8);
            kprintf("sceIoLseek addr: %08X\n", (u32)sceIoLseek_func);
        }
        pspSdkSetK1(k1);
        SceOff cur = sceIoLseek_func(fd, 0, PSP_SEEK_CUR);
        SceOff size = sceIoLseek_func(fd, 0, PSP_SEEK_END);
        sceIoLseek_func(fd, cur, PSP_SEEK_SET);
        int k1 = pspSdkSetK1(0);
        kprintf("DLC already decrypted for fd: %08X, returning %u for sceNpDrmEdataGetDataSize\n", fd, (u32)size);
        pspSdkSetK1(k1);
        return size;
    }
    pspSdkSetK1(k1);
    SceOff size = sceNpDrmEdataGetDataSize_func(fd);
    pspSdkSetK1(0);
    kprintf("sceNpDrmEdataGetDataSize called, fd: %08X, size: %u\n", fd, (u32)size);
    pspSdkSetK1(k1);
    return size;
}

// checks if the edat/sprx file hasn't been renamed, must return 0 on success (sceNpDrmRenameCheck)
SceUID np_rename(const char *file) {
    int k1 = pspSdkSetK1(0);
    if (!sceNpDrmRenameCheck_func) {
        sceNpDrmRenameCheck_func = (void *) sctrlHENFindFunction("scePspNpDrm_Driver", "scePspNpDrm_user", 0x275987D1);
        kprintf("sceNpDrmRenameCheck addr: %08X\n", (u32)sceNpDrmRenameCheck_func);
    }
    int res = check_encryption(file);
    switch (res) {
    case 0:
        kprintf("DLC already decrypted for %s, returning 0 for sceNpDrmRenameCheck\n", file);
        break;
    default:
        kprintf("Failed to open/read %s, code: %08X\n", file, res);
    case 1:
        pspSdkSetK1(k1);
        res = sceNpDrmRenameCheck_func(file);
        k1 = pspSdkSetK1(0);
        kprintf("sceNpDrmRenameCheck called, file: %s, res: %08X\n", file, res);
    }
    pspSdkSetK1(k1);
    return res;
}

// (sceKernelLoadModule2?) special function to launch sprx files from the memory stick
SceUID np_load(const char *path, int flags, SceKernelLMOption *option) {
    int k1 = pspSdkSetK1(0);
    if (!sceKernelLoadModule2_func) {
        sceKernelLoadModule2_func = (void *) sctrlHENFindFunction("sceModuleManager", "ModuleMgrForUser", 0xF2D8D1B4);
        kprintf("Found sceKernelLoadModule2 addr: %08X\n", (u32)sceKernelLoadModule2_func);
    }
    if (!sceKernelLoadModule_func) {
        sceKernelLoadModule_func = (void *) sctrlHENFindFunction("sceModuleManager", "ModuleMgrForUser", 0x977DE386);
        kprintf("Found sceKernelLoadModule addr: %08X\n", (u32)sceKernelLoadModule_func);
    }
    SceUID res = -1;
    switch (check_encryption(path)) {
    case 0: // DECRYPTED
        kprintf("DLC already decrypted for %s, using sceKernelLoadModule\n", path);
        pspSdkSetK1(k1);
        res = sceKernelLoadModule_func(path, flags, option);
        pspSdkSetK1(0);
        kprintf("sceKernelLoadModule called, path: %s, flags: %08X, option: %08X, modid: %08X\n", path, flags, (u32)option, res);
        break;
    case 1: // ENCRYPTED
        kprintf("DLC encrypted for %s, using sceKernelLoadModule2\n", path);
        pspSdkSetK1(k1);
        res = sceKernelLoadModule2_func(path, flags, option);
        pspSdkSetK1(0);
        kprintf("sceKernelLoadModule2 called, path: %s, flags: %08X, option: %08X, modid: %08X\n", path, flags, (u32)option, res);
        break;
    default:
        kprintf("Error in check_encryption!\n");
    }
    pspSdkSetK1(k1);
    return res;
}
